/*--------------------------------------------------------------*
 * Morse Code to Eprom Hex File Program cw2hex			*
 * Copyright G. Forrest Cook W0RIO (formerly WB0RIO)- 1996-2020	*
 *								*
 * This code has been released under the			*
 * GNU GENERAL PUBLIC LICENSE version 2: http://www.gnu.org/	*
 *								*
 *  Version History:						*
 *  v1.2.2 10/25/2013 Extra characters added			*
 *  v1.2.3 7/15/2015 Fixed hex file checksum bug		*
 *  v1.2.4 7/15/2015 Added FSCW signal on bit 2			*
 *  v1.2.5 7/17/2015 Changed timing after letters and spaces	*
 *   to 3/7, reorganized character table and improved		*
 *   documentation.						*
 *  v1.2.6 7/20/2015 Added -c option for CSV file output and	*
 *   added support for the tx enable bit.			*
 *  v1.2.7 8/3/2015 rearranged the output bits			*
 *  v1.2.8 9/27/2016 added FSCW toggle				*
 *  v1.2.9 9/12/2018 added E S (and) code			*
 *  v1.3.0 9/17/2018 Made ASCII to Morse translation a table.	*
 *  v1.3.1 3/13/2020 Added (SOS) char, improved comments.	*
 *								*
 * An example beacon circuit that works with this program is	*
 *  here: http://www.solorb.com/elect/hamcirc/beacon/		*
 *								*
 * This program takes text input and converts it to Morse	*
 * encoded characters in the form of a standard			*
 * Intel I8HEX format HEX file.					*
 * The Hex file output can be used to program an EPROM which	*
 * can then be used in a simple circuit which generates		*
 * a repeating CW message.  Possible uses are CW beacon		*
 * transmitters and CQ calling machines.			*
 *								*
 * If you start the program with the -c option, it will produce	*
 * a csv output file which can be used to build an array for an	*
 * Arduino microcontroller beacon program or other C software.	*
 *								*
 * The 1989 ARRL Handbook page 19-2 documents some of the	*
 * more obscure Morse code characters, many of them are 	*
 * implemented in this program. Wikipedia also has a useful	*
 * description of Morse code cnd defines a few more unusual	*
 * Morse characters.						*
 *								*
 * The output data uses the lower four bits as follows:		*
 * The D0 output bit is the keyed morse code line.		*
 * The D1 bit has keyed Frequency Shift CW (FSCW) data,		*
 *  optionally use this to shift the frequency of all dashes.	*
 * The D2 bit is the TX Enable line.				*
 * The D3 output bit is used to reset a counter in continuous	*
 * loop mode or to stop a counter in single message mode.	*
 *								*
 * The input string for a typical beacon message might be:	*
 * ~ >> {W0RIO/B GS DM79IX~<<<<					*
 * The tilde starts the oscillator, the space allows the	*
 * oscillator to stabilize, the >> produces a			*
 * steady tone followed by (DE) CALLSIGN/B Grid Square Data,	*
 * the second tilde turns the oscillator off and the <<<<	*
 * produces a silent pause after the message has been sent.	*
 *								*
 *--------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>

#define HEXFILENAME "cw.hex"
#define CSVFILENAME "cw.csv"
#define FSCWBIT 1		/* output bit 1 */
#define FSCWMASK 0xFD		/* and mask to turn off FSCW bit */
#define TXENABLE 4		/* output bit 2 */
#define STOPBIT 8		/* output bit 3 */

/*
 * function declarations
 */
int outhexbyte();
int outcsvbyte();
void hexbout();
void hexwout();
int csumfinalize();

/*
 * Global variables
 */
FILE *Outfile=0;	/* Output file pointer */
int CSVout=0;		/* CSV (Comma Separated Values) output flag */


int main (int argc, char **argv)
{
	char *cp, ch, instr[500];
	int i, j, len, opt;
	int outval;
	int txenable = 0;
	int fscwenable = 0;

/*
 * This table translates 7 bit ASCII to characters in the morse table.
 * The first 32 control characters are omitted, 0 is null (unused).
 */
	static int ascii2morse[96]=
	{
	  27,38,39,40,41,42,43,44,45,46,47,48,49,50,51,52, /* 32-47 */
	  28,29,30,31,32,33,34,35,36,37,53,54,63,55,64,56, /* 48-63 0-9 */
	  60,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,	   /* 64-79 A-Z */
	  16,17,18,19,20,21,22,23,24,25,26,0,59,0,0,61,	   /* 80-95 */
	  62,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,	   /* 96-111 a-Z */
	  16,17,18,19,20,21,22,23,24,25,26,58,57,0,0,0	   /* 112-127 */
	};

/*
 * This table contains the encoded Morse code characters.
 *
 * In each output character, bit 0 is the keying bit and bit 1 is the
 * frequency shift bit, bit 1 is high only during dashes.
 * In other words: dots are 1, and dashes are 3,3,3,
 * Inter element gaps and trailing spaces are 0, 
 * A 9 ends each character definition, it is not transmitted.
 */
	static unsigned char morse[65][27]=
	{
	  {9},						/* null character */
	  {1,0,3,3,3,0,0,0,9},					/* A [1] */
	  {3,3,3,0,1,0,1,0,1,0,0,0,9},
	  {3,3,3,0,1,0,3,3,3,0,1,0,0,0,9},
	  {3,3,3,0,1,0,1,0,0,0,9},
	  {1,0,0,0,9},
	  {1,0,1,0,3,3,3,0,1,0,0,0,9},
	  {3,3,3,0,3,3,3,0,1,0,0,0,9},
	  {1,0,1,0,1,0,1,0,0,0,9},
	  {1,0,1,0,0,0,9},
	  {1,0,3,3,3,0,3,3,3,0,3,3,3,0,0,0,9},
	  {3,3,3,0,1,0,3,3,3,0,0,0,9},
	  {1,0,3,3,3,0,1,0,1,0,0,0,9},
	  {3,3,3,0,3,3,3,0,0,0,9},
	  {3,3,3,0,1,0,0,0,9},
	  {3,3,3,0,3,3,3,0,3,3,3,0,0,0,9},
	  {1,0,3,3,3,0,3,3,3,0,1,0,0,0,9},
	  {3,3,3,0,3,3,3,0,1,0,3,3,3,0,0,0,9},
	  {1,0,3,3,3,0,1,0,0,0,9},
	  {1,0,1,0,1,0,0,0,9},
	  {3,3,3,0,0,0,9},
	  {1,0,1,0,3,3,3,0,0,0,9},
	  {1,0,1,0,1,0,3,3,3,0,0,0,9},
	  {1,0,3,3,3,0,3,3,3,0,0,0,9},
	  {3,3,3,0,1,0,1,0,3,3,3,0,0,0,9},
	  {3,3,3,0,1,0,3,3,3,0,3,3,3,0,0,0,9},
	  {3,3,3,0,3,3,3,0,1,0,1,0,0,0,9},			/* Z [26] */
	  {0,0,0,0,0,0,0,9},					/* space */
	  {3,3,3,0,3,3,3,0,3,3,3,0,3,3,3,0,3,3,3,0,0,0,9},	/* 0 [28] */
	  {1,0,3,3,3,0,3,3,3,0,3,3,3,0,3,3,3,0,0,0,9},		/* 1 */
	  {1,0,1,0,3,3,3,0,3,3,3,0,3,3,3,0,0,0,9},		/* 2 */
	  {1,0,1,0,1,0,3,3,3,0,3,3,3,0,0,0,9},			/* 3 */
	  {1,0,1,0,1,0,1,0,3,3,3,0,0,0,9},			/* 4 */
	  {1,0,1,0,1,0,1,0,1,0,0,0,9},				/* 3 */
	  {3,3,3,0,1,0,1,0,1,0,1,0,0,0,9},			/* 5 */
	  {3,3,3,0,3,3,3,0,1,0,1,0,1,0,0,0,9},			/* 7 */
	  {3,3,3,0,3,3,3,0,3,3,3,0,1,0,1,0,0,0,9},		/* 8 */
	  {3,3,3,0,3,3,3,0,3,3,3,0,3,3,3,0,1,0,0,0,9},		/* 9 [37] */
	  {3,3,3,0,1,0,3,3,3,0,1,0,3,3,3,0,3,3,3,0,0,0,9},	/* ! KW  */
	  {1,0,3,3,3,0,1,0,1,0,3,3,3,0,1,0,0,0,9},		/* " AF [39] */
	  {1,0,1,0,1,0,3,3,3,0,1,0,0,0,9},			/* ack # SN */
	  {1,0,1,0,1,0,3,3,3,0,1,0,1,0,3,3,3,0,0,0,9},		/* $ SX */
	  {1,0,3,3,3,0,1,0,3,3,3,0,1,0,1,0,0,0,9},	     /* parag. % AL */
	  {1,0,3,3,3,0,1,0,1,0,1,0,0,0,9},			/* & AS */
	  {1,0,3,3,3,0,3,3,3,0,3,3,3,0,3,3,3,0,1,0,0,0,9},	/* ' WG */
	  {3,3,3,0,1,0,3,3,3,0,3,3,3,0,1,0,0,0,9},		/* ( KN [45] */
	  {3,3,3,0,1,0,3,3,3,0,3,3,3,0,3,3,3,0,1,0,0,0,9},	/* ) KG */
	  {3,3,3,0,1,0,3,3,3,0,1,0,3,3,3,0,0,0,9},	  /* attention * KA */
	  {1,0,3,3,3,0,1,0,3,3,3,0,1,0,0,0,9},			/* eom + AR */
	  {3,3,3,0,3,3,3,0,1,0,1,0,3,3,3,0,3,3,3,0,0,0,9},	/* , GW */
	  {3,3,3,0,1,0,1,0,1,0,1,0,3,3,3,0,0,0,9},		/* - DU [50] */
	  {1,0,3,3,3,0,1,0,3,3,3,0,1,0,3,3,3,0,0,0,9},		/* . RK */
	  {3,3,3,0,1,0,1,0,3,3,3,0,1,0,0,0,9},			/* / DN */
	  {3,3,3,0,3,3,3,0,3,3,3,0,1,0,1,0,1,0,0,0,9},		/* : OS */
	  {3,3,3,0,1,0,3,3,3,0,1,0,3,3,3,0,1,0,0,0,9},		/* ; KR */
	  {3,3,3,0,1,0,1,0,1,0,3,3,3,0,0,0,9},			/* = BT [55] */
	  {1,0,1,0,3,3,3,0,3,3,3,0,1,0,1,0,0,0,9},		/* ? UD */
	  {1,0,1,0,1,0,3,3,3,0,1,0,3,3,3,0,0,0,9},		/* end | SK */
	  {3,3,3,0,1,0,1,0,0,1,0,0,0,9},			/* ID { D E */
	  {1,0,0,1,0,1,0,1,0,0,0,9},				/* \ E S [59] */
	  {1,0,3,3,3,0,3,3,3,0,1,0,3,3,3,0,1,0,0,0,9},		/* @ WR */
	  {1,0,1,0,3,3,3,0,3,3,3,0,1,0,3,3,3,0,0,0,9},		/* _ UK */
	  {1,0,1,0,1,0,3,3,3,0,3,3,3,0,3,3,3,0,1,0,1,0,1,0,0,0,9}, /* ` SOS */
	  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,9},		/* < pause */
	  {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,9}		/* > tone[64] */
	};

	/*
	 * Print the startup message.
	 */
	printf ("\ncw2hex V1.3.1 - W0RIO CW to Intel Hex file encoder");
	printf ("\nThis is free and open-source software, License: GPLv2.");
	printf ("\nEnter cw2hex -c to output a cw.csv file.");
	printf ("\n\nValid Chars: a-z, A-Z, 0-9, space and special characters:");
	printf ("\nEnter      Send          Enter      Send");
	printf ("\n-----      ----          -----      ----");
	printf ("\n\\ and      (E S)         \"          (AF)");
	printf ("\n{ id       (D E)         \'          (WG)");
	printf ("\n%% parag.   (AL)          $          (SX)");
	printf ("\n# ack      (SN)          (          (KN)");
	printf ("\n* start    (KA)          )          (KG)");
	printf ("\n+ eom      (AR)          ,          (GW)");
	printf ("\n| end      (SK)          -          (DU)");
	printf ("\n= break    (BT)          .          (RK)");
	printf ("\n& wait     (AS)          :          (OS)");
	printf ("\n?          (UD)          ;          (KR)");
	printf ("\n/          (DN)          !          (KW)");
	printf ("\n@          (WR)          _          (UK)");
	printf ("\n`          (SOS)");
	printf ("\n<  Pause 20 Units        >  Tone 20 Units");
	printf ("\n~ Toggle TX Enable       ^ Toggle FSCW Mode");

	/*
	 * Check for command line arguments.
	 */
        while ((opt = getopt (argc, argv, "c")) != EOF)
        {
                switch (opt)
                {
                /*
                 * csv output format.
                 */
                    case 'c':
			CSVout = 1;
                        break;
                /*
                 * Other stuff.
                 */
                    default:
                        printf ("Unknown option: %c\n", opt);
                        break;
                }
        }


/*
 * Open the output file.
 */
	if (CSVout)	/* csv format */
	{
		if((Outfile = fopen(CSVFILENAME, "w")) == NULL)
		{
			printf("cw2hex: can't open output file %s\n",
					CSVFILENAME);
			exit(1);
		}
	}
	else		/* hex format */
	{
		if((Outfile = fopen(HEXFILENAME, "w")) == NULL)
		{
			printf("cw2hex: can't open output file %s\n",
					HEXFILENAME);
			exit(1);
		}
	}

	cp=instr;	/* point to beginning of the input string */

/*
 * Get the string to encode here.
 */
	printf ("\n\nType a message to encode: ");
	fgets (instr, sizeof(instr), stdin);

/*
 * Loop through all of the characters of the input string.
 */
	while (*cp)
	{
		ch=(*cp++) & 0x7F;		/* make sure char is 7 bits */

		if (ch == '\n')			/* end on newline */
			break;
		else if (ch >= 32)		/* ignore control chars */
		{

/*
 * Process special mode change characters.
 */
		    if (ch == '~') /* tilde */
		    {
			if (txenable)
			{
				txenable=0;
				printf ("\nTx Enable Off");
			}
			else
			{
				txenable = TXENABLE;
				printf ("\nTx Enable On");
			}
		    }
		    else if (ch == '^') /* carrot */
		    {
			if (fscwenable)
			{
				fscwenable=0;
				printf ("\nFSCW Off");
			}
			else
			{
				fscwenable = 1;
				printf ("\nFSCW On");
			}
		    }

		    else	/* Process the normal Morse chars */
		    {
		      i=ascii2morse[(int)ch-32]; /* offset for control chars */
		      printf ("\n(%c) ", ch);

		      j=0;

		      while (morse[i][j] != 9)
		      {
			if (fscwenable)
				outval=(int)morse[i][j++]+txenable;
			else
				outval=(int)(morse[i][j++]&FSCWMASK)+txenable;

			if (CSVout)
				outcsvbyte (outval,0);
			else
				outhexbyte (outval,0);
		      }
		    }
		}
	}

	printf ("\n(stop) ");

/*
 * Get the final length of the file then close up shop.
 */
	if (CSVout)
	{
		len=outcsvbyte (STOPBIT,0);	/* output stop bit character */
		outcsvbyte (0,1);		/* finalize the csv file */
	}
	else
	{
		len=outhexbyte (STOPBIT,0);	/* output stop bit character */
		outhexbyte (0,1);		/* finalize the hex file */
	}

	fclose (Outfile);

	printf ("\n\nThis message is %d bytes long.", len);

	if (CSVout)
		printf ("\nThe output file is called %s\n\n", CSVFILENAME);
	else
		printf ("\nThe output file is called %s\n\n", HEXFILENAME);

	return (-1);
}

/*--------------------------------------------------------------*/

/*
 * Output a CSV format file.
 */

#define CLINEBYTES 10

int outcsvbyte (int byte, int finish)
{
	static int bcount=0;
	static int addr=0;

	if (!finish)	/* output a byte */
	{
		printf ("%01X", byte);

		fprintf (Outfile, "0x");	/* print one data element */
		hexbout (byte);

		if (bcount < CLINEBYTES-1)	/* not last byte in line */
		{
			fprintf (Outfile, ", ");
			bcount++;
		}
		else				/* last byte in line */
		{
			fprintf (Outfile, ",\n");
			bcount=0;
		}
		addr++;
	}
	else	/* finish */
	{
		fprintf (Outfile, "\n\n%d\n", addr);	/* Byte Count */

		return (0);
	}

	return (addr);
}

/*--------------------------------------------------------------*/

/*
 * Output bytes to an Intel Hex format file.
 */

#define HLINEBYTES 16

int outhexbyte (int byte, int finish)
{
	static int bcount=0;
	static int firstline=1;
	static int chksum=HLINEBYTES; /* first part of checksum is byte len. */
	static int addr=0;
	int i;

	if (firstline)
	{
		firstline=0;
		fprintf (Outfile, ":10000000"); /* Write beginning of file. */
	}

	if (!finish)	/* output a byte */
	{
		printf ("%01X", byte);

		if (bcount < HLINEBYTES)	/* not last byte in line */
			hexbout (byte);		/* output one hex byte */
		else	/* end of line, output checksum and start new line */
		{
			hexbout (csumfinalize(chksum));

			bcount=0;

			fprintf (Outfile, "\n:");

			hexbout (HLINEBYTES);
			chksum=HLINEBYTES;

			hexwout (addr);
			chksum += addr>>8;
			chksum += addr&255;

			fprintf (Outfile, "00");	/* pad byte */

			hexbout (byte);
		}
		bcount++;
		chksum += byte;
		addr++;
	}
	else	/* Fill out the last line with 00 bytes. */
	{
		for (i=1; i<=HLINEBYTES-bcount; i++)
		{
			hexbout (0);
			addr++;
		}

		hexbout (csumfinalize(chksum));

		fprintf (Outfile, "\n:00000001FF\n");	/* write EOF line */

		return (0);
	}

	return (addr);
}

/*--------------------------------------------------------------*/

void hexbout(byte)	/* output a byte in hex */
int     byte;
{
	char    *hexstr = { "0123456789ABCDEF" } ;

	byte = byte & 0377;
	fprintf (Outfile,"%c%c",hexstr[byte>>4],hexstr[byte&017]);
}

/*--------------------------------------------------------------*/

void hexwout(word)	/* output a word in hex */
int     word;
{
        hexbout(word>>8);	/* high byte of word */
	hexbout(word&255);	/* low byte of word */
}

/*--------------------------------------------------------------*/

int csumfinalize(dval)	/* calculate final checksum value */
int     dval;
{
	int rval;

	dval &= 255;			/* chop off incoming high bits */

	rval = (~dval + 1) & 255;	/* twos complement and mask */

	return (rval);
}

/*--------------------------------------------------------------*/
